function stimulus = make_stimulus(parameters)
    
    % clean up the parameters
    
    if nargin
        parameters = missing_params(parameters);
    else
        parameters = default_stimulus_parameters();
    end

    % figure out foreground/background labeling
    
    fg_region = determine_foreground(parameters);
    
    % get the textures for foreground and background
    
    [fg_texture, bg_texture] = make_textures(parameters);

    % combine labeling and textures
    
    stimulus = fg_texture .* fg_region + bg_texture .* (1 - fg_region);
end


%
%  determine_foreground
%

function fg_region = determine_foreground(s)
%%

[x,y] = meshgrid(1:s.image_size, 1:s.image_size);
xs = repmat(x, [1 1 s.n_frames]);
ys = repmat(y, [1 1 s.n_frames]);

switch s.type
    
    case 'foreground_only'
        c0 = [0.5 0.5] * s.image_size + 1;
        r = s.image_size;
        boundary_velocity = [0 0];
        fn = in_moving_area (@in_square, c0, r, boundary_velocity, ...
                 s.image_size, s.image_size, s.n_frames);
        fg_region = fn(xs, ys);
        
    case 'aperture'
        r = 0.25;
        c0 = ceil([0.5 0.5] * s.image_size);
        r = s.image_size * r;
        fn = in_moving_area (@in_circle, c0, r, s.boundary_velocity, ...
                 s.image_size, s.image_size, s.n_frames);
        fg_region = fn(xs, ys);
        
    case 'michotte'
        r = 0.20;
        c0 = ceil([r 0.5] * s.image_size);
        r = s.image_size * r;
        fn = in_moving_area (@in_circle, c0, r, s.boundary_velocity, ...
                 s.image_size, s.image_size, s.n_frames);
        fg_region = fn(xs, ys) & (xs < ceil(0.55 * s.image_size));
        
    case 'kaplan'
%       c0 = [-0.1 0.5] * s.image_size + 1;
        c0 = [-0.1 0.5] * s.image_size + 1 - 0.5 * s.boundary_velocity * s.n_frames;
        r = 0.6 * s.image_size;
        fn = in_moving_area (@in_square, c0, r, s.boundary_velocity, ...
                 s.image_size, s.image_size, s.n_frames);
        fg_region = round(fn(xs, ys));
        
    case 'top'
        c0 = [0.5 -0.1] * s.image_size + 1 - 0.5 * s.boundary_velocity * s.n_frames;
        r = 0.6 * s.image_size;
        fn = in_moving_area (@in_square, c0, r, s.boundary_velocity, ...
                 s.image_size, s.image_size, s.n_frames);
        fg_region = round(fn(xs, ys));
        
    case 'royden'
        % Doesn't support motion yet
        aspect_ratio = s.royden.aspect_ratio;
        unit_side = 0.4 * s.image_size; % length of square when aspect_ratio=1
        r = [unit_side * aspect_ratio, unit_side / aspect_ratio] / 2;
        fn = in_moving_area (@in_rect, 0.5 * s.image_size * [1 1], ...
                             r, s.boundary_velocity, s.image_size, ...
                             s.image_size, s.n_frames);
        fg_region = fn(xs, ys);
        
    case 'koenderink'
        fg_region = zeros(size(xs));
        stripe_r = 0.125 * s.image_size;
        stripe_spacing = 0.5 * s.image_size;
        [cx0, cy0] = spaced_grid(s.image_size, s.image_size, s.n_frames, ...
                                 [stripe_spacing s.image_size], ...
                                 s.boundary_velocity);
        r = [stripe_r, s.image_size];
        for j = 1:numel(cx0)
            fn = in_moving_area(@in_rect, [cx0(j) cy0(j)], [r r], ...
                   s.boundary_velocity, s.image_size, s.image_size, s.n_frames);
            fg_region = min(fg_region + fn(xs, ys), 1);
        end
        
    otherwise
        disp('Unknown stimulus type')
        
end
end


function [fg_texture, bg_texture] = make_textures(s)
%%

[x,y] = meshgrid(1:s.image_size, 1:s.image_size);
xs = repmat(x, [1 1 s.n_frames]);
ys = repmat(y, [1 1 s.n_frames]);

for i = {'fg', 'bg'}
    
    texture = zeros(s.image_size, s.image_size, s.n_frames);
    texture_type = eval(['s.' i{1} '_texture']);
    v = eval(['s.' i{1} '_velocity']);
    
    switch texture_type
        
        case 'white'
            texture = ones(size(xs));
            
        case 'gray'
            texture = 0.5 * ones(size(xs));
            
        case 'black'
            texture = zeros(size(xs));
            
        case 'checkerboard'
            [cx0, cy0] = spaced_grid(s.image_size, s.image_size, s.n_frames, ...
                                      2 * s.dot_size * [1 1], v);
            cx0 = floor([cx0(:); cx0(:) + s.dot_size]);
            cy0 = floor([cy0(:); cy0(:) + s.dot_size]);

            for j = 1:numel(cx0)
                fn = in_moving_area(@in_square, [cx0(j) cy0(j)], ...
                         ceil((s.dot_size - 1) / 2), round(v), ...
                         s.image_size, s.image_size, s.n_frames);
                texture = min(texture + fn(xs, ys), 1);
            end
            
        case '+'
        case 'X'
            
        case 'stripes'
            period = eval(['s.' i{1} '_period']);
            pattern = zeros(1, period);
            pattern(1 : period/2) = 1;
            for y = 1 : s.image_size;
                for t = 1: s.n_frames;
                    texture(:,y,t) = pattern(1 + mod(y - v(1) * t, period));
                end
            end
            
        case 'ramp'
            period = eval(['s.' i{1} '_period']);
            pattern = [0 : period] / period;
            for y = 1 : s.image_size;
                for t = 1: s.n_frames;
                    texture(:,y,t) = pattern(1 + mod(y - v(1) * t, period + 1));
                end
            end
%           texture = xs / s.image_size;
            
        case 'sawtooth'
            period = eval(['s.' i{1} '_period']);
            pattern = min([1 : period] - 1, [period : -1 : 1]);
            pattern = pattern / max(pattern);
            for y = 1 : s.image_size;
                for t = 1: s.n_frames;
                    texture(:,y,t) = pattern(1 + mod(y - v(1) * t, period));
                end
            end
            
        case 'sinusoid'
            period = eval(['s.' i{1} '_period']);
            w = 1 : period;
            pattern = 0.5 + 0.5 * sin(2 * pi * (w - 1) / period);
            for x = 1 : s.image_size;
                for y = 1 : s.image_size;
                    for t = 1: s.n_frames;
                        w = mod(y - v(1) * t, period);
                        c = x / s.image_size;
                        texture(x,y,t) = 0.5 + 0.5 * c * sin(2 * pi * w / period);
                    end
                end
            end
            
        case 'grayscale'
            vel = [v(2) v(1)];
            siz = s.image_size + abs(vel) * (s.n_frames - 1);
            pattern = rand(siz);
            contrast = 0.2;
            pattern = min(max(contrast * randn(siz) + 0.5, 0), 1);
            origin = 1 + max(vel, 0) * (s.n_frames - 1);
            
            for t = 1: s.n_frames;
                pos = origin - vel * (t - 1);
                texture(:,:,t) = pattern(pos(1) : pos(1) + s.image_size - 1, pos(2) : pos(2) + s.image_size - 1);
            end
            
        case 'grayscale-flicker'
            
            % velocity is irrelevant
            
            siz = s.image_size;
            contrast = 0.2;
            
            for t = 1: s.n_frames;
                pattern = min(max(contrast * randn(siz) + 0.5, 0), 1);
                texture(:,:,t) = pattern(:,:);
            end
            
        case 'random'
            
            handle = eval(['@in_' s.dot_shape]);
            
            [cx0, cy0] = spaced_grid(s.image_size, s.image_size, s.n_frames, ...
                                     s.dot_spacing * [1 1], v);
            cx0 = cx0 + randn(size(cx0)) * s.dot_spacing / 6;
            cy0 = cy0 + randn(size(cy0)) * s.dot_spacing / 6;
            
            for j = 1:numel(cx0)
                fn = in_moving_area(handle, [cx0(j) cy0(j)], s.dot_size, ...
                         v, s.image_size, s.image_size, s.n_frames);
                texture = min(texture + fn(xs, ys), 1);
            end
            
        case 'dense'
            handle = eval(['@in_' s.dot_shape]);
            [cx0, cy0] = spaced_grid(s.image_size, s.image_size, s.n_frames, ...
                                     [1 1], v);
            [tl,br] = window_size(s.image_size, s.image_size, ...
                                  s.n_frames, v);
            rsiz = br - tl + 1;
            rsiz_odd = floor(rsiz / 2) * 2 + 1;
            [x,y] = meshgrid(linspace(-1,1,max(rsiz_odd)));
            mask = ifftshift(x.^2 + y.^2 >= 0.5^2); % hi-pass
            tex  = ifft2(mask .* fft2(rand(max(rsiz_odd))));
            tex  = tex(1:rsiz(1), 1:rsiz(2));
            for j = 1:numel(cx0)
                fn = in_moving_area(handle, [cx0(j) cy0(j)], 0.0001, ...
                         round(v), s.image_size, s.image_size, s.n_frames);
%                 texture = min(texture + fn(xs, ys), 1);
                texture = min(texture + (tex(j) > 0) * fn(xs, ys), 1);
            end
            % NOTE: DOESN'T SUPPORT MOTION YET (bitmap-ish, others are
            % vector-ish, this will take some hacking
%             texture = repmat(rand(s.image_size), [1 1 s.n_frames]);

        case 'dense-flicker'
            
            % velocity is irrelevant
            
            handle = eval(['@in_' s.dot_shape]);
            
            [cx0, cy0] = spaced_grid(s.image_size, s.image_size, s.n_frames, ...
                                     [1 1], v);
            [tl,br] = window_size(s.image_size, s.image_size, ...
                                  s.n_frames, v);
            rsiz = br - tl + 1;
            rsiz_odd = floor(rsiz / 2) * 2 + 1;
            [x,y] = meshgrid(linspace(-1,1,max(rsiz_odd)));
            mask = ifftshift(x.^2 + y.^2 >= 0.5^2); % hi-pass
            tex  = ifft2(mask .* fft2(rand(max(rsiz_odd))));
            tex  = tex(1:rsiz(1), 1:rsiz(2));
            for j = 1:numel(cx0)
                fn = in_moving_area(handle, [cx0(j) cy0(j)], 0.0001, ...
                         round(v), s.image_size, s.image_size, s.n_frames);
%                 texture = min(texture + fn(xs, ys), 1);
                texture = min(texture + (tex(j) > 0) * fn(xs, ys), 1);
            end
            % NOTE: DOESN'T SUPPORT MOTION YET (bitmap-ish, others are
            % vector-ish, this will take some hacking
%             texture = repmat(rand(s.image_size), [1 1 s.n_frames]);

        case 'checker'
%             [cx0, cy0] = spaced_grid(s.image_size, s.image_size, s.n_frames, ...
%                                      2 * s.dot_size * [1 1], v);
%             cx0 = floor([cx0(:); cx0(:) + s.dot_size]);
%             cy0 = floor([cy0(:); cy0(:) + s.dot_size]);
            [cx0, cy0] = spaced_grid(s.image_size, s.image_size, s.n_frames, ...
                                     [10 10], v);
            cx0 = floor(cx0(:)) + 0.5;
            cy0 = floor(cy0(:)) + 0.5;
            cx0 = [cx0(:); cx0(:) + 5];
            cy0 = [cy0(:); cy0(:) + 5];
            for j = 1:numel(cx0)
%                 fn = in_moving_area(@in_square, [cx0(j) cy0(j)], ...
%                          ceil((s.dot_size - 1) / 2), round(v), ...
%                          s.image_size, s.image_size, s.n_frames);
                fn = in_moving_area(@in_square, [cx0(j) cy0(j)], ...
                         2.4, round(v), ...
                         s.image_size, s.image_size, s.n_frames);
                texture = min(texture + fn(xs, ys), 1);
            end
            
        case 'sparse_regular'
            handle = eval(['@in_' s.dot_shape]);
            [cx0, cy0] = spaced_grid(s.image_size, s.image_size, s.n_frames, ...
                                     s.sparse_regular.spacing * [1 1], v);
            cx0 = floor(cx0 + 2);
            cy0 = floor(cy0 + 2);
            for j = 1:numel(cx0)
                fn = in_moving_area(handle, [cx0(j) cy0(j)], 0.0001, ...
                         round(v), s.image_size, s.image_size, s.n_frames);
                texture = min(texture + fn(xs, ys), 1);
            end
            
        case 'twoframe'
            v = round(v);
            texture = zeros(s.image_size, s.image_size, s.n_frames);
            n_dots_per_frame = round(s.twoframe.density / 2 * (s.image_size^2));
            for j = 1:(s.n_frames - 1)
                center = ceil(s.image_size * rand(n_dots_per_frame, 2));
                indices = sub2ind(size(texture), ...
                    min(max(center(:, 1), 1), s.image_size), ...
                    min(max(center(:, 2), 1), s.image_size), ...
                                  j * ones(n_dots_per_frame, 1));
                texture(indices) = 1;
                indices = sub2ind(size(texture), ...
                    min(max(center(:, 1) + v(2), 1), s.image_size), ...
                    min(max(center(:, 2) + v(1), 1), s.image_size), ...
                                  (j + 1) * ones(n_dots_per_frame, 1));
                texture(indices) = 1;
            end
            
        otherwise
            disp('whu?')
    end
    
    eval([i{1} '_texture = texture;'])
end
end


function [top_left, bottom_right] = window_size (nx, ny, nt, v)
top_left     = [1   1] - nt * max( v, [0 0]);
bottom_right = [nx ny] + nt * max(-v, [0 0]);
end


function [cx0, cy0] = spaced_grid (nx, ny, nt, spacing, v)

[top_left, bottom_right] = window_size(nx, ny, nt, v);
[cx0, cy0] = meshgrid(top_left(1):spacing(1):bottom_right(1), ...
                      top_left(2):spacing(2):bottom_right(2));
cx0 = cx0 + 0.5 * (bottom_right(1) - cx0(end));
cy0 = cy0 + 0.5 * (bottom_right(2) - cy0(end));
end


function fn = in_moving_area (handle, c0, r, v, nx, ny, nt)
pos = ((1:nt)' * v + ones(nt, 1) * c0)';
cx = shiftdim(repmat(ones(ny, 1) * pos(1, :), [1 1 nx]), 2);
cy = shiftdim(repmat(ones(ny, 1) * pos(2, :), [1 1 nx]), 2);
fn = @(x,y) handle(x, y, cx, cy, r);
end


function z = in_circle (x, y, cx, cy, r)
z = alias((x - cx) .^ 2 + (y - cy) .^ 2 - r .^ 2);
end


function z = in_diamond (x, y, cx, cy, r)
if numel(r) == 1
    r(2) = r(1);
end
z = alias((abs(x - cx) - r(1)) + (abs(y - cy) - r(2)));
end


function z = in_rect (x, y, cx, cy, r)
if numel(r) == 1
    r(2) = r(1);
end
z = alias(max(abs(x - cx) - r(1), abs(y - cy) - r(2)));
end


function z = in_square (x, y, cx, cy, r)
z = in_rect (x, y, cx, cy, [r r]);
end


function z = alias(w)
z = min(max(1 - w, 0), 1);
end

%
%  default = missing_params(input_parameters)
%

function default = missing_params(input_parameters)

    default = default_stimulus_parameters;
    
    % Fill in missing parameters with default values
    
    names = fieldnames(default);
    
    for i = 1 : numel(names)
        if isfield(input_parameters, names{i})
            default.(names{i}) = input_parameters.(names{i});
        end
    end
end
